"""
Base API
"""

import requests
from .exceptions import ParameterError, ApiKeyError, ApiFailure
from .utils import snake_to_camel_case

SYSTRANIO_URL = 'https://api-platform.systran.net'


class BaseAPI(object):
    """
    Base API class
    Subclasses all systran.io APIs (Translation, NLP, etc.)
    Calls are made with `requests`
    """
    key = ''

    def __init__(self, key):
        """
        Sets the API Key
        Get yours on https://systran.io
        """
        self.key = key

    def _set_attributes(self, **kwargs):
        """
        Loops through kwargs and tries to set attributes
        Attributes are API parameters so if it fails we raise an exception
        """
        for parameter, value in kwargs.items():
            try:
                getattr(self, parameter)  # trigger error
                setattr(self, parameter, value)
            except AttributeError:
                raise ParameterError(
                    'Invalid API parameter : {}'.format(parameter))

    @property
    def payload(self):
        """
        Like __dict__() but with camelCase cases to send them to systran as parameters
        """
        output = {}
        for key, value in self.__dict__.items():
            if not key.startswith('_'):  # ignore private
                output[snake_to_camel_case(key)] = value
        return output

    def _call_url(self, endpoint):
        """
        Returns the full call url: host + endpoint
        """
        return '{}{}'.format(SYSTRANIO_URL, endpoint)

    def _response(self, response):
        """
        Handles requests' response.
        Returns the json response or raises an exception
        """
        if response.status_code == 200:
            return response.json()
        if response.status_code == 500:
            try:
                result = response.json()
                raise ApiFailure(result['error']['message'])
            except ValueError:
                raise ApiFailure('Unknown error')
        if response.status_code == 400:
            raise ApiKeyError('Invalid API key : {}'.format(self.key))

    def post(self, endpoint):
        """
        Sends a POST request to systran with all the parameters
        previously set.
        """
        return self._response(
            requests.post(self._call_url(endpoint), data=self.payload))

    def get(self, endpoint):
        """
        Sends a GET request to systran with all the parameters
        previously set. Returns the json response or raises an exception
        """
        return self._response(
            requests.get(self._call_url(endpoint), data=self.payload))



"""
Translation API

Usage :

```
import systranio

translation = systranio.Translation(API_KEY)
options = {'source': 'en' }
result = translation.text('traduction', 'fr', **options)
print(result)  # traduction
print(result.stats)  #  {"elapsed_time": 20, "nb_characters": 9, ...
```

"""

from .api import BaseAPI

TEXT_ENDPOINT = '/translation/text/translate'
API_VERSION_ENDPOINT = '/translation/apiVersion'
SUPPORTED_LANGUAGES_ENDPOINT = '/translation/supportedLanguages'


class TextTranslationResult(object):
    """
    Takes a text translation result (json) and makes it usable
    repr as the translation output
    """

    output = None
    detected_language = None
    detected_language_confidence = 0
    stats = {}

    def __init__(self, result: dict):
        """
        `output` is always there, but not the others
        So we try to politely set the values
        """
        self.output = result.get('output')
        self.detected_language = result.get('detectedLanguage')
        self.detected_language_confidence = result.get(
            'detectedLanguageConfidence')
        self.stats = result.get('stats')

    def __repr__(self) -> str:
        """
        The translated string
        """
        return self.output


class LanguagePairResult(object):
    """
    Takes a supported_languages result and makes it usable
    """

    source = None
    target = None
    profiles = {}

    def __init__(self, result: dict):
        """
        we try to politely set the values
        """
        self.source = result.get('source')
        self.target = result.get('target')
        self.profiles = result.get('profiles')

    def __str__(self) -> str:
        """
        A representation of the language pair : `source` → `target`
        """
        return '{} → {}'.format(self.source, self.target)


class Translation(BaseAPI):
    """
    Translation API model
    """

    input = ''
    target = ''
    source = 'auto'
    format = None
    profile = None
    with_source = None
    with_annotations = False
    with_dictionary = None
    with_corpus = None
    back_translation = False
    options = None
    encoding = 'utf-8'
    callback = None

    def text(self, text: str, target: str, **kwargs) -> TextTranslationResult:
        """
        Translates a single `text` into `target` language
        and returns a TextResult object

        TODO
        * handle multiple input (and loop result['outputs'])
        """
        self.input = text
        self.target = target
        self._set_attributes(**kwargs)
        response = self.post(TEXT_ENDPOINT)
        return TextTranslationResult(response['outputs'][0])

    def api_version(self) -> str:
        """
        Current version for translation apis
        """
        self._set_attributes()
        response = self.get(API_VERSION_ENDPOINT)
        return response['version']

    def supported_languages(self, source=None, target=None) -> list:
        """
        List of language pairs in which translation is supported.
        This list can be limited to a specific source language or target language.
        """
        self.source = source
        self.target = target
        results = []
        response = self.get(SUPPORTED_LANGUAGES_ENDPOINT)
        for pair in response['languagePairs']:
            results.append(LanguagePairResult(pair))
        return results



"""
Translation unit tests
"""

import os
import unittest

import systranio
from systranio.exceptions import ParameterError, ApiKeyError, ApiFailure
from systranio.Translation import TextTranslationResult, LanguagePairResult

API_KEY = os.environ['SYSTRANIO_KEY']  # required


class TestTranslation(unittest.TestCase):
    """
    Test Translation API
    """

    def setUp(self):
        self.translation = systranio.Translation(API_KEY)

    def test_invalid_api_key(self):
        "Failure with a missing api key"
        translation = systranio.Translation('')
        with self.assertRaises(ApiKeyError):
            translation.text('pas de clé !', 'en')

    def test_invalid_parameter(self):
        "Invalid parameter added to the translation query"
        options = {'jean_michel': 'margaret'}
        with self.assertRaises(ParameterError):
            self.translation.text('cats', 'en', **options)

    def test_api_failure(self):
        "An invalid route (translation from en to en)"
        options = {'source': 'en'}
        with self.assertRaises(ApiFailure):
            self.translation.text('what ?', 'en', **options)

    def test_simple_text_translation(self):
        "A simple translation must returns something"
        options = {'source': 'en'}
        result = self.translation.text('hello', 'fr', **options)
        self.assertIsNotNone(result.output)

    def test_text_translation_returns(self):
        "A simple translation returns TextTranslationResult"
        options = {'source': 'en'}
        result = self.translation.text('hello', 'fr', **options)
        self.assertIsInstance(result, TextTranslationResult)

    def test_api_version(self):
        "api version number should be 1.0.0"
        version = self.translation.api_version()
        self.assertAlmostEqual(version, '1.0.0')

    def test_supported_languages_list(self):
        "supported_language returns a list"
        languages = self.translation.supported_languages()
        self.assertIsInstance(languages, list)

    def test_supported_languages_item(self):
        "supported_language items  a list of LanguagePairResult"
        languages = self.translation.supported_languages()
        self.assertIsInstance(languages[0], LanguagePairResult)


class TestTextTranslationResult(unittest.TestCase):
    """
    Test TextTranslationResult
    """

    def setUp(self):
        self.api_response = {
            'output': 'diffamation',
            'stats': {
                'nb_characters': 10,
                'nb_tus_failed': 0,
                'nb_tus': 1,
                'nb_tokens': 1,
                'elapsed_time': 20
            }
        }

    def test_repr(self):
        "__repr__ should be output"
        result = TextTranslationResult(self.api_response)
        self.assertEqual(str(result), result.output)


class TestLanguagePairResult(unittest.TestCase):
    """
    Test LanguagePairResult
    """

    def setUp(self):
        self.api_response = {
            "source": "nl",
            "target": "en",
            "profiles": [{
                "id": 0,
                "private": False
            }]
        }

    def test_str(self):
        "__repr__ should be output"
        result = LanguagePairResult(self.api_response)
        self.assertEqual(str(result), 'nl → en')


if __name__ == '__main__':
    unittest.main()



"""
Setup script
"""
from setuptools import setup, find_packages

REQUIRES = ['requests >= 2']

README = open('README.md').read()

setup(
    name='systranio',
    version='0.1.3',  # managed by bumbversion
    description='A simple REST API client for Systran.io',
    author='Canarduck',
    author_email='renaud@canarduck.com',
    url='https://github.com/canarduck/systran-translation-client',
    keywords=['Systran', 'Rest Translation API'],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description=README,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha', 'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Linguistic'
    ], )



"""
Tools
"""

def snake_to_camel_case(snake):
    """
    Converts a snake_case string to camelCase
    @see: https://codereview.stackexchange.com/a/85314
    """
    first, *rest = snake.split('_')
    return first + ''.join(word.capitalize() for word in rest)



"""
Exceptions
"""

class ParameterError(AttributeError):
    """
    Invalid API Parameter set
    """


class ApiKeyError(ValueError):
    """
    Systran.io refused the key provided
    """

class ApiFailure(Exception):
    """
    Systran.io returned a 500 error
    """
